from data import loadimage, loadmusic, loadfont, loadsound
from pygame import Surface, Rect, mixer, key, BLEND_RGBA_MULT
from pygame import K_SPACE, K_s, K_DOWN, K_p
from constants import RESOLUTION_X, RESOLUTION_Y, SILLY_STUFF, OBSTACLES
from constants import INFORMATION_BITES, ECTION_BITES
from constants import INFORMATION_THANKS, ECTION_THANKS, UNDEFLECTIONS
from random import random, choice as rchoice, randint
from math import floor
from interface import Text
from script import script1, script2, script3

# size of little player char
sprite_w = 80
sprite_h = 138

# y of the player status information
# measured from the halfway point
status_y = 45

# padding / y positions for interview text
pad = 7
first = pad
second = pad + 25
third = pad + 50

# the dialogue script "format"
interviewer = 0
player = 1

bg_width = RESOLUTION_X

bgs = [
    "background1.jpg",
    "background2.jpg",
    "background3.jpg",
    "background4.jpg"
]

obs_colorses = [
    [
        [69, 0, 3, 255],
        [92, 0, 2, 255],
        [148, 9, 13, 255]
    ],
    [
        [242, 71, 56, 255],
        [242, 89, 75, 255],
        [242, 131, 107, 255]
    ]
]

green_colors = [
    [40, 153, 118, 255],
    [83, 184, 222],
    [177, 255, 145, 255]
]


white = [255, 255, 255, 255]
black = [0, 0, 0, 255]
blacks = [
    [10, 10, 10, 255],
    [30, 30, 30, 255],
    [50, 50, 50, 255]
]
damage = [255, 35, 35, 255]


class PlayStage(object):

    level_confs = [
        {
            "song":  "level2_music.ogg",
            "obs_freq": .1,
            "script": script1,
            "health": 3,
            "length": 45,
            "biteoffset": 0
        },
        {
            "song": "level1_music.ogg",
            "obs_freq": .3,
            "script": script2,
            "health": 5,
            "length": 130,
            "biteoffset": 2
        },
        {
            "song": "level1_music.ogg",
            "obs_freq": .5,
            "script": script3,
            "health": 5,
            "length": 200,
            "biteoffset": 4
        }
    ]

    def __init__(self, game):
        self.game = game
        level_idx = self.game.next_game_level
        level_conf = self.level_confs[level_idx]

        self.top = Television(game, level_conf)
        self.bottom = RunnerGame(game, level_conf)

        self.intro = True

        loadmusic(level_conf["song"])
        mixer.music.play(-1)

    def update(self, dt):
        self.top.update(dt)
        self.bottom.update(dt)

        if self.bottom.winning:
            self.top.winning = True
        if self.bottom.losing:
            self.top.losing = True

        if self.top.start_game:
            self.bottom.waiting = False

    def draw(self, surf):
        self.top.draw(surf)
        self.bottom.draw(surf)


class Dialog(object):

    def __init__(self, game, dialog):
        self.game = game
        self.player_char_name = "Ms. Information"
        if self.game.player_char > 0:
            self.player_char_name = "Mr. Ection"
        self.dialog = []
        for d in dialog:
            if len(d) < 2:
                self.dialog.append(d)
            else:
                self.dialog.append(
                    [
                        d[0],
                        d[1].replace("<player>", self.player_char_name)
                    ]
                )

        self.start_game = False
        self.curr_script_idx = 0
        self.curr_line = 0
        self.prev_char = -1
        self.curr_char = 0
        self.size = 17
        self.currently_speaking = interviewer
        self.char_timer = 0
        self.line_timer = 0
        self.lines = [
            Text("",
                 font="OpenSans-CondBold.ttf",
                 size=self.size, pos=(pad, first)),
            Text("",
                 font="OpenSans-CondBold.ttf",
                 size=self.size, pos=(pad, second)),
            Text("",
                 font="OpenSans-CondBold.ttf",
                 size=self.size, pos=(pad, third))
        ]
        self.set_color()
        for l in self.lines:
            l.boring = True
            l.hilite = True

        self.dialog_over = False
        self.losing = False

    def advancescript(self):
        self.curr_script_idx += 1
        if self.curr_script_idx >= len(self.dialog):
            self.dialog_over = True
            return

        scriptline = self.dialog[self.curr_script_idx]
        self.currently_speaking = scriptline[0]
        self.set_color()
        self.curr_char = 0
        self.prev_char = -1
        self.curr_line = 0
        self.line_timer = 0
        self.char_timer = 0
        for l in self.lines:
            l.clear()

    def deflecting(self, text):
        self.dialog = [
            [1, text],
            [0, rchoice(UNDEFLECTIONS)],
            [3]
        ]
        self.curr_script_idx = -1
        self.advancescript()
        self.dialog_over = False

    def lose_end(self):
        self.curr_script_idx = -1
        self.advancescript()
        self.dialog_over = False

    def update(self, dt):

        # if self.dialog_over:
        #    return

        scriptline = self.dialog[self.curr_script_idx]
        """
        if len(scriptline) < 2:
            command = scriptline[0]
            print("Command: {}".format(command))
            if command == 2:
                self.start_game = True
            self.advancescript()
        """
        try:
            next_d = self.dialog[self.curr_script_idx + 1]
            if len(next_d) < 2:
                if next_d[0] == 2:
                    self.dialog_over = True
                    self.start_game = True
                if next_d[0] == 3:
                    self.dialog_over = True
        except:
            pass

        if self.curr_char >= len(scriptline[1]):
            self.line_timer += dt * .8
            if self.line_timer > 8 and not self.dialog_over:
                self.advancescript()
                return
            else:
                return

        if self.curr_char > self.prev_char:
            next_line = scriptline[1]
            next_char = next_line[self.curr_char]
            # skip space on first char of new line
            if next_char == " " and len(self.lines[self.curr_line].text) < 1:
                self.curr_char += 1
                self.prev_char += 1
                next_char = next_line[self.curr_char]

            self.lines[self.curr_line].append(
                next_line[self.curr_char].upper())
            self.prev_char += 1

        maxw = RESOLUTION_X - ((pad * 2) + 10)
        if self.lines[self.curr_line].line_width > maxw:
            if self.curr_line < 2:
                self.curr_line += 1
            elif self.curr_line == 2:
                self.advanceline()

        self.char_timer += dt * 12
        advancechar = len(scriptline) > 1 and \
            self.curr_char < len(scriptline[1])
        if self.char_timer > 3 and advancechar:
            self.curr_char += 1
            self.char_timer = 0

        for t in self.lines:
            t.update(dt)

    def set_color(self):
        for l in self.lines:
            if self.game.player_char == 0:
                if self.currently_speaking == interviewer:
                    l.colors = [[186, 212, 255, 255]]
                else:
                    l.colors = [[255, 186, 186, 255]]
            else:
                if self.currently_speaking == interviewer:
                    l.colors = [[255, 186, 186, 255]]
                else:
                    l.colors = [[186, 212, 255, 255]]

    def advanceline(self):
        self.lines.pop(0)
        self.lines.append(
            Text("",
                 font="OpenSans-CondBold.ttf",
                 size=self.size, pos=(pad, third)))
        self.lines[2].boring = True
        self.lines[2].hilite = True
        self.lines[0].pos = (pad, first)
        self.lines[1].pos = (pad, second)
        self.set_color()

    def draw(self, surf):
        for t in self.lines:
            t.draw(surf)


class TalkingHead(object):

    def __init__(self, head, pos, game):
        self.game = game
        self.head = loadimage(head).convert_alpha()
        self.talking = False
        self.frameoffset = 0
        self.frames = [0, 0, 0, 1, 0, 0, 1, 1]
        self.pos = pos

    def update(self, dt):
        if self.talking:
            self.frameoffset += dt * 2
            if self.frameoffset >= len(self.frames):
                self.frameoffset = 0
        else:
            self.frameoffset = 0

    def draw(self, surf):
        f_offset = self.frames[floor(self.frameoffset)] * 148
        src = Rect(f_offset, 0, 148, 196)
        surf.blit(self.head, self.pos, src)


class Television(object):

    def __init__(self, game, conf):
        self.game = game
        self.conf = conf
        self.tcolor = [255, 255, 255, 255]
        self.tfmt = "{}  -  "
        self.bg = loadimage("interview_background.png").convert_alpha()

        self.ticker_font = loadfont("OpenSans-CondBold.ttf", 11)
        self.tickers = [rchoice(SILLY_STUFF) for _ in range(1, 5)]
        self.ticker_surfs = [self.ticker_font.render(
            self._text(x), True, self.tcolor
        ) for x in self.tickers]
        self.ticker_widths = [self.ticker_font.size(
            self._text(x))[0] for x in self.tickers]
        self.ticker_offset = 0
        self.start_game = False

        self.dialog = Dialog(game, conf["script"])

        self.soundbite_timer = 0
        self.next_soundbite = 30  # randint(100, 300)
        self.soundbite_index = conf["biteoffset"]
        self.info_bites = [(loadsound(x[0]), x[1]) for x in INFORMATION_BITES]
        self.ection_bites = [(loadsound(x[0]), x[1]) for x in ECTION_BITES]
        self.curr_channel = None
        self.winning = False
        self.losing = False

        h1pos = (30, 107)
        h2pos = (236, 107)

        if self.game.player_char == 0:
            self.interviewer_head = TalkingHead(
                "mrection_head.png", h1pos, game)
            self.player_head = TalkingHead(
                "msinformation_head.png", h2pos, game)
        else:
            self.interviewer_head = TalkingHead(
                "msinformation_head.png", h1pos, game)
            self.player_head = TalkingHead(
                "mrection_head.png", h2pos, game)

    def play_soundbite(self):
        bites = None
        if self.game.player_char == 0:
            bites = self.info_bites
        else:
            bites = self.ection_bites

        self.soundbite_index += 1
        if self.soundbite_index >= len(self.info_bites):
            self.soundbite_index = 0
        self.curr_channel = bites[self.soundbite_index][0].play()
        self.dialog.deflecting(bites[self.soundbite_index][1])

    def _text(self, t):
        return self.tfmt.format(t.upper())

    def update(self, dt):
        self.ticker_offset += dt * 8
        if self.ticker_offset > self.ticker_widths[0]:
            self.tickers.pop(0)
            self.ticker_surfs.pop(0)
            self.ticker_widths.pop(0)
            self.tickers.append(rchoice(SILLY_STUFF))
            self.ticker_widths.append(
                self.ticker_font.size(
                    self._text(self.tickers[-1]))[0])
            self.ticker_surfs.append(
                self.ticker_font.render(
                    self._text(self.tickers[-1]), False, self.tcolor))
            self.ticker_offset = 0
        self.dialog.update(dt)
        if not self.dialog.dialog_over:
            if self.dialog.currently_speaking == 0:
                self.interviewer_head.talking = True
                self.player_head.talking = False
            else:
                self.interviewer_head.talking = False
                self.player_head.talking = True

        self.interviewer_head.update(dt)

        self.start_game = self.dialog.start_game

        if self.start_game and not self.winning and not self.losing:
            self.soundbite_timer += dt * .4
            if self.soundbite_timer > self.next_soundbite:
                self.soundbite_timer = 0
                self.next_soundbite = randint(80, 150)
                self.play_soundbite()

        if self.curr_channel is not None:
            if self.curr_channel.get_busy():
                self.player_head.talking = True
            else:
                self.player_head.talking = False
        else:
            self.player_head.talking = False

        if self.losing:
            if not self.dialog.losing:
                i = 0
                for s in self.conf['script']:
                    if s[0] == 2:
                        break
                    i += 1
                self.dialog.dialog = self.conf['script'][i+1:]
                self.dialog.losing = True
                self.dialog.lose_end()
            elif self.dialog.dialog_over:
                self.game.state.lose_game()

        self.player_head.update(dt)

    def draw(self, surf):
        # draw bg
        surf.blit(self.bg, (0, 0))
        # draw ticker
        accum = 0
        for w, s in zip(self.ticker_widths, self.ticker_surfs):
            x = accum - self.ticker_offset
            if x > RESOLUTION_X:
                continue
            surf.blit(s, (x, 345))
            accum += w
        # draw dialog
        self.dialog.draw(surf)
        self.interviewer_head.draw(surf)
        self.player_head.draw(surf)


class Obstacle:

    def __init__(self, pos, word, y, speed):
        self.speed = speed
        if y == 0:
            y = 200
        else:
            y = 300
        self.rect = Rect(pos, y, 30, 39)
        self.text = Text(
            word,
            size=23, pos=(self.rect.x, self.rect.y),
            font="VarelaRound-Regular.ttf")
        self.text.colors = rchoice(obs_colorses)
        self.text.wavespeed = .2
        self.text.hilite = True
        self.scored = False
        self.hit = False

    def update(self, dt):
        self.rect.left -= dt * self.speed
        self.text.pos = (self.rect.x, self.rect.y)
        self.text.update(dt)

    def draw(self, surf):
        self.text.draw(surf)
        # surf.fill([244, 10, 10, 255], self.rect)


class ScorePoint:
    def __init__(self, points, streak):
        size = min(12 + streak, 40)
        self.text = Text("+{}".format(points), size=size, pos=(90, 230))
        self.life = 40

    def update(self, dt):
        self.text.pos = (
            self.text.pos[0],
            min(self.text.pos[1] - (10 * dt), 200)
        )
        self.text.colors = green_colors
        self.text.hilite = True
        self.text.update(dt)
        self.life -= dt * 3

    def draw(self, surf):
        self.text.draw(surf)


class HealthIndicator:
    def __init__(self, health):
        self.text = Text(
            "♥" * health, font="OpenSansEmoji.ttf", pos=(320, status_y))
        self.label = Text("HEALTH", pos=(280, status_y))
        self.value = health
        self.text.colors = obs_colorses[1]
        self.text.hilite = True

    def hit(self):
        self.value -= 1
        if self.value > -1:
            self.text.shorten_by_1()

    def update(self, dt):
        self.text.update(dt)
        self.label.update(dt)

    def draw(self, surf):
        self.text.draw(surf)
        self.label.draw(surf)


class RunnerGame(object):

    def __init__(self, game, conf):
        self.game = game

        self.score = 0
        self.points = []
        self.streak = 0
        self.game.last_score = 0
        self.game.last_streak = 0
        self.longest_streak = 0
        self.was_never_hit = True

        self.freq = conf["obs_freq"]

        self.length = conf['length']
        self.time_left = self.length
        self.prev_time = self.time_left
        self.score_label = self.make_text("SCORE:", (30, status_y))
        self.time_label = self.make_text("TIME LEFT:", (150, status_y))
        self.score_text = self.make_text(self.score, (80, status_y))
        self.time_text = self.make_text(self.time_left, (220, status_y))

        self.surf = Surface((RESOLUTION_X, RESOLUTION_Y / 2.0)).convert_alpha()
        self.walls = [0, 0, 0, 255]

        if self.game.player_char == 0:
            self.char = RunnerChar(game, "msinformation_sprites.png")
            self.win_sound = loadsound(ECTION_THANKS)
        else:
            self.char = RunnerChar(game, "mrection_sprites.png")
            self.win_sound = loadsound(INFORMATION_THANKS)

        self.bgs = [loadimage(x).convert() for x in bgs]
        self.curr_bgs = [rchoice(self.bgs) for _ in range(1, 3)]

        self.ready_text = Text(
            "GET READY...", size=40, pos=(40, 100),
            hilite=True, font="VarelaRound-Regular.ttf")

        self.go_text = Text(
            "GO!!!", size=40, pos=(80, 100),
            hilite=True, font="VarelaRound-Regular.ttf")

        self.go_text.colors = green_colors
        self.go_text.speed = 20

        self.win_text = Text(
            "NICE!!", size=90, pos=(40, 80),
            hilite=True, font="VarelaRound-Regular.ttf")
        self.win_text.colors = green_colors

        self.lose_text = Text(
            "OUCH!!",  size=90, pos=(40, 80),
            hilite=True, font="VarelaRound-Regular.ttf")
        self.lose_text.colors = rchoice(obs_colorses)

        self.waiting = True

        self.timer = 0
        self.distance = 0

        self.started = False
        self.start_time = 0

        self.winning = False
        self.win_timer = 0
        self.losing = False

        self.speed = 20
        self.bgoffset = 0

        self.obstacles = []

        self.hit_sound = loadsound("fail.ogg")
        self.dodge_sound = loadsound("goodjob.ogg")

        self.health = conf["health"]
        self.health_indicator = HealthIndicator(self.health)

    def score_points(self):
        val = 10 + self.streak * 10
        self.score += val
        self.streak += 1
        if self.streak > self.longest_streak:
            self.longest_streak = self.streak
        self.points.append(ScorePoint(val, self.streak))
        self.dodge_sound.play()
        self.score_text = self.make_text(self.score, (80, status_y))

    def make_text(self, t, pos):
        text = Text(str(t), pos=pos)
        text.colors = blacks
        text.hilite = True
        text.wavespeed = .2
        text.amplitude = .5
        return text

    def update(self, dt):
        self.timer += dt

        if self.waiting:
            self.ready_text.update(dt)
        else:
            if not self.started:
                self.started = True
                self.start_time = self.timer

        diepoints = []
        for i, p in enumerate(self.points):
            p.update(dt)
            if p.life < 0:
                diepoints.append(i)

        for p in diepoints:
            self.points.pop(p)

        self.health_indicator.update(dt)
        self.score_label.update(dt)
        self.score_text.update(dt)
        self.time_label.update(dt)
        self.time_text.update(dt)

        if self.start_time < self.timer < self.start_time + 20:
            self.go_text.update(dt)

        if self.losing:
            self.char.hit = True
            self.lose_text.update(dt)
            self.game.last_score = self.score
            self.game.last_streak = self.longest_streak
            return

        self.bgoffset -= dt * self.speed
        if self.bgoffset < -bg_width:
            self.bgoffset = 0
            self.curr_bgs.pop(0)
            self.curr_bgs.append(rchoice(self.bgs))

        self.char.update(dt)

        if self.waiting:
            return

        if self.winning:
            self.win_text.update(dt)
            self.obstacles = []
            self.win_timer += dt * 1
            self.game.last_score = self.score
            self.game.last_streak = self.longest_streak
            if self.win_timer > 25:
                self.game.state.win_game()
            return

        if self.started:
            self.time_left -= dt * .1
            t = floor(self.time_left)
            if self.prev_time > t:
                self.prev_time = t
                self.time_text = self.make_text(self.prev_time, (220, status_y))

        if self.time_left < 0:
            self.win_sound.play()
            self.winning = True

        if len(self.obstacles) == 0:
            self.add_obs(RESOLUTION_X + 150)
        elif len(self.obstacles) < 3:
            obs_end = self.obstacles[-1].rect.x + self.obstacles[-1].rect.width
            fwd_x = max(
                (random() * 1000) - (self.freq * 1000),
                obs_end + (400 * (1 - self.freq)))
            self.add_obs(fwd_x)

        hit = False
        obstacles_to_remove = []
        for i, o in enumerate(self.obstacles):
            o.update(dt)
            if self.char.collide_rect.colliderect(o) and not o.scored:
                hit = True
                self.char.hit = True
                self.streak = 0
                o.hit = True
                obstacles_to_remove.append(i)
                self.hit_sound.play()
                self.char.debug_color = [255, 0, 0, 255]
                self.health_indicator.hit()
                self.health -= 1

        if not hit:
            self.char.debug_color = [0, 255, 0, 255]
        else:
            for o in obstacles_to_remove:
                self.obstacles.pop(o)

        dodgepoint = self.char.char_pos[0] + (sprite_w / 2.0)
        if self.obstacles[0].rect.x < dodgepoint:
            if not self.obstacles[0].scored and not self.obstacles[0].hit:
                self.score_points()
                self.obstacles[0].scored = True

        if self.obstacles[0].rect.x < -self.obstacles[0].text.line_width - 10:
            self.obstacles.pop(0)

        if self.health < 1:
            self.losing = True

    def add_obs(self, pos):
        self.obstacles.append(
                Obstacle(pos, rchoice(OBSTACLES), round(random()), self.speed))

    def draw(self, surf):
        self.surf.fill([0, 0, 0, 0])

        self.surf.blit(self.curr_bgs[0], (self.bgoffset, 0))
        self.surf.blit(self.curr_bgs[1], (self.bgoffset + bg_width, 0))

        if self.waiting:
            self.ready_text.draw(self.surf)
        else:
            if self.start_time < self.timer < self.start_time + 20:
                self.go_text.draw(self.surf)

        for o in self.obstacles:
            if o.rect.x < RESOLUTION_X:
                o.draw(self.surf)

        self.char.draw(self.surf)

        for p in self.points:
            p.draw(self.surf)

        self.health_indicator.draw(self.surf)
        self.score_label.draw(self.surf)
        self.score_text.draw(self.surf)
        self.time_label.draw(self.surf)
        self.time_text.draw(self.surf)

        if self.losing:
            self.lose_text.draw(self.surf)
        if self.winning:
            self.win_text.draw(self.surf)

        surf.blit(self.surf, (0, RESOLUTION_Y / 2.0))


class RunnerChar:

    def __init__(self, game, sprite):
        self.game = game
        self.char = loadimage(sprite).convert_alpha()
        self.char_pos = (50, 200)
        self.collide_rect = Rect(
            self.char_pos[0] + sprite_w / 4.0,
            self.char_pos[1],
            sprite_w / 2.0,
            sprite_h
        )

        self.start_jump = False

        self.jumping = False
        self.jump_time = 0

        self.landed = False
        self.landed_time = 0

        self.hit = False
        self.hit_timer = 0

        self.crouching = False
        self.walkstep = 0
        self.walksteps = [0, 1, 0, 2]
        self.base_color = white

    def update(self, dt):

        self.walkstep += dt * 1.2
        if self.walkstep >= 4:
            self.walkstep = 0

        keys = key.get_pressed()

        if keys[K_SPACE]:
            self.start_jump = True
        else:
            if self.start_jump is True:
                self.jumping = True
            self.start_jump = False

        if self.jumping:
            self.jump_time += dt * .4
            self.collide_rect.height = sprite_h - 50
            if self.jump_time > 2.6:
                self.jumping = False
                self.landed = True
                self.jump_time = 0
        elif keys[K_DOWN] or keys[K_s]:
            self.crouching = True
            self.collide_rect.height = sprite_h - 50
            self.collide_rect.y = self.char_pos[1] + 50
        else:
            self.crouching = False
            self.collide_rect.height = sprite_h
            self.collide_rect.y = self.char_pos[1]

        if self.landed:
            self.landed_time += dt * .4
            if self.landed_time > .9:
                self.landed = False
                self.landed_time = 0

        if self.hit:
            self.hit_timer += dt * .5
            if self.hit_timer > 2:
                self.hit = False
                self.hit_timer = 0
            self.base_color = damage
        else:
            self.base_color = white

        if keys[K_p]:
            self.game.state.pause()

    def draw(self, surf):

        src = None
        show_crouch = self.start_jump or self.landed
        if not self.jumping and not self.crouching and not show_crouch:
            f_offset = self.walksteps[floor(self.walkstep)] * sprite_w
            src = Rect(f_offset, 0, sprite_w, sprite_h)
        elif show_crouch:
            src = Rect(3 * sprite_w, 0, sprite_w, sprite_h)
        elif self.jumping:
            src = Rect(4 * sprite_w, 0, sprite_w, sprite_h)
        elif self.crouching:
            src = Rect(5 * sprite_w, 0, sprite_w, sprite_h)

        s = self.char.copy()
        csurf = self.char.copy()
        csurf.fill(self.base_color)
        s.blit(csurf, (0, 0), None, BLEND_RGBA_MULT)

        surf.blit(
            s,
            self.char_pos,
            src
        )

        # surf.fill(self.debug_color, self.collide_rect, BLEND_MAX)
